{
rmCommutator V 1.00 - component for Delphi.

rmCommutator is a component for Delphi,
written in Delphi 3,
modified in Delphi 4 and 5.

Compatibility:
Delphi 1 : not supported without changes in code
Delphi 2 : certainely works but not tested
Delphi 3 : tested, OK
Delphi 4 : tested, OK
Delphi 5 : tested, OK

------------------------------------------------------------
Description
------------------------------------------------------------
This component is designed to simulate multi-positions switch.
It can be multi-position, as a multi-point commutator.
It can have multiple lines, as matrix style.

As it can also be single, with only the two positions On or Off,
I suggest the use of rmSwitch component, specialy designed
for simulate single slide-switch or push-button.


------------------------------------------------------------
Actually supported functions
------------------------------------------------------------

AutoSize
--------
When you set AutoSize property to True, Width and Height properties
have no effect. Width and height of rmCommutator component are determined
by the properties listed bellow:

BorderMargin    - space between rmCommutator border and switch border
ExtremityMargin - space between rmCommutator extremity and first / last switch
SwitchLength    - switch size: height if rmCommutator is in vertical orientation
SwitchSpacing   - space between every switch
SwitchWidth     - switch size: width if rmCommutator is in vertical orientation
RowCount        - number of switches by line
LineCount       - number of lines of switches

When you set AutoSize property to false (default value), Width and Height
properties determine width and height of rmCommutator component.

Orientation property lets you chose orientation (ah ?).


------------------------------------------------------------
Actually no supported functions (i work on it for next release)
------------------------------------------------------------
property MultiSelect (Multiple switches selection) with associated
selCount property.


------------------------------------------------------------
Distribution and copyrights
------------------------------------------------------------
As i take initial ideas of multiple sources codes
(of public domain on internet and CD-ROM) to create this
component, i think it's normal to release it to public domain.

Thanks to all peoples who freely distribute their works (sometimes
hardly obtained).

Je sais, j'aurais pu joindre  ce code source un texte
en francais !


------------------------------------------------------------
Any suggestions, comments, or bugs report
------------------------------------------------------------
MALLARD Remy
FRANCE
remy.mallard@wanadoo.fr


------------------------------------------------------------
Component History
------------------------------------------------------------
11/04/2007
Version : 1.0

Actually no bug known

}


unit rmCommutator;

interface

uses
  WinTypes, WinProcs, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs,  ExtCtrls,
  rmBase,   // for basic classes (for exemple TrmBorderStyle classe)
  rmoeled,  // for led class
  rmoeSwitch;  // for switch class

type
  TOrientation = (voVertical, voHorizontal);
  //TBorder3DStyle = (csRaised, csNone, csLowered);
  //TLedBorderStyle = (lsRaised, lsNone, lsLowered);
  //TSwitchStyle = (ssNone, ssAllSwitches, ssOnlyActiveSwitch);

  TrmCommutator = class;


{------------------------------------------------------------------------}
{   TSwitchOn and TSwitchOff, derived from TrmBasicShape (rmBase.pas unit)
{------------------------------------------------------------------------}

  TSwitchOn = class(TrmBasicShape)
  published
    property Shape;
    property BorderStyle;
    property BorderWidth;
    property BorderColor;
    property ActiveColor;
    property HilightColor;
    property ShaddowColor;
  end;

  TSwitchOff = class(TrmBasicShape)
  published
    property Shape;
    property BorderStyle;
    property BorderWidth;
    property BorderColor;
    property InactiveColor;
    property HilightColor;
    property ShaddowColor;
  end;


{------------------------------------------------------------------------}
{   TSwitchLedOn and TSwitchLedOff, derived from TrmBasicShape (rmBase.pas unit)
{------------------------------------------------------------------------}

  TSwitchLedOn = class(TrmBasicShape)
  published
    property Shape;
    property Width;
    property Height;
    property BorderStyle;
    property BorderWidth;
    property BorderColor;
    property ActiveColor;
    property HilightColor;
    property ShaddowColor;
  end;

  TSwitchLedOff = class(TrmBasicShape)
  published
    property Shape;
    property Width;
    property Height;
    property BorderStyle;
    property BorderWidth;
    property BorderColor;
    property InactiveColor;
    property HilightColor;
    property ShaddowColor;
  end;


{------------------------------------------------------------------------}
{   TrmCommutator
{------------------------------------------------------------------------}

  TrmCommutator = class(TGraphicControl)
  private
    FBmpBuffer: TBitmap;
    MouseCapt: boolean;
    FAllowAllUp: boolean;
    FShowLed: boolean;
    FLedOn: TSwitchLedOn;
    FLedOff: TSwitchLedOff;
    FSwitchOn: TSwitchOn;
    FSwitchOff: TSwitchOff;
    FBGColor: TColor;
    FBorderStyle: TrmBorderStyle;
    Forientation: TOrientation;
    FMultiSelect: boolean;
    FBorderMargin: integer;
    FAutoSize: boolean;
    FRowCount: integer;
    FLineCount: integer;
    FSwitchIndex: integer;
    FSwitchSpacing: integer;
    FSwitchWidth: integer;
    FSwitchHeight: integer;
    FOnChange: TNotifyEvent;
    procedure SetAllowAllUp(value: boolean);
    procedure SetShowLed(value: boolean);
    procedure SetBGColor(value: TColor);
    procedure SetBorderStyle(value: TrmBorderStyle);
    procedure SetOrientation(value: TOrientation);
    procedure SetBorderMargin(value: integer);
    procedure SetAutoSize(value: boolean);
    procedure SetRowCount(value: integer);
    procedure SetLineCount(value: integer);
    procedure SetSwitchIndex(value: integer);
    procedure SetSwitchWidth(value: integer);
    procedure SetSwitchHeight(value: integer);
    procedure SetSwitchSpacing(value: integer);
  protected
    procedure Paint; override;
    procedure UpdateBmpBuffer;
    procedure GetSelectedSwitch(X, Y: integer);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
  public
    SelectedRow,
    SelectedLine: integer;
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
    procedure Free;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  published
    property AllowAllUp: boolean read FAllowAllUp write SetAllowAllUp;
    property ShowLed: boolean read FShowLed write SetShowLed;
    property LedOn: TSwitchLedOn read FLedOn write FLedOn;
    property LedOff: TSwitchLedOff read FLedOff write FLedOff;
    property SwitchOn: TSwitchOn read FSwitchOn write FSwitchOn;
    property SwitchOff: TSwitchOff read FSwitchOff write FSwitchOff;
    property BGColor: TColor read FBGColor write SetBGColor;
    property BorderStyle: TrmBorderStyle read FBorderStyle write SetBorderStyle;
    property Orientation: TOrientation read FOrientation write SetOrientation;
    property BorderMargin: integer read FBorderMargin write SetBorderMargin;
    property AutoSize: boolean read FAutoSize write SetAutoSize;
    property RowCount: integer read FRowCount write SetRowCount;
    property LineCount: integer read FLineCount write SetLineCount;
    property SwitchIndex: integer read FSwitchIndex write SetSwitchIndex;
    property SwitchWidth: integer read FSwitchWidth write SetSwitchWidth;
    property SwitchHeight: integer read FSwitchHeight write SetSwitchHeight;
    property Spacing: integer read FSwitchSpacing write SetSwitchSpacing;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    { Standard properties and events}
    property Height;
    property Width;
    property PopupMenu;
    property Hint;
    property ShowHint;
    property Tag;
    property Visible;
  end;

  procedure Register;


implementation

procedure Register;
begin
  RegisterComponents('OptoElec', [TrmCommutator]);
end;

{------------------------------------------------------------------------}
{   TrmCommutator
{------------------------------------------------------------------------}

constructor TrmCommutator.Create(Owner: TComponent);
begin
  Inherited Create(Owner);
//  Parent := owner;
  FBmpBuffer := TBitmap.Create;
  FLedOn := TSwitchLedOn.create(self);
  FLedOff := TSwitchLedOff.create(self);
  FSwitchOn := TSwitchOn.create(self);
  FSwitchOff := TSwitchOff.create(self);
  Height := 18;
  Width := 194;
  FShowLed := true;
  MouseCapture := false;
  FAllowAllUp := false;
  FBGColor := clGray;
  FOrientation := voHorizontal;
  FBorderStyle := bsLowered;
  FBorderMargin := 2;
  FAutoSize := true;
  FRowCount := 12;
  FLineCount := 1;
  FSwitchIndex := 0;
  FSwitchWidth := 14;
  FSwitchHeight := 14;
  FSwitchSpacing := 2;
  with FLedOn do begin
    ActiveColor := clRed;
    Active := true;
    Width := 4;
    Height := 4;
    Shape := stEllipse;
    BorderStyle := bsNone;
  end;
  with FLedOff do begin
    Active := false;
    InactiveColor := clAqua;
    Width := 2;
    Height := 2;
    Shape := stEllipse;
    BorderStyle := bsNone;
  end;
  With FSwitchOn do begin
    Active := true;
    Shape := stRectangle;
    FSwitchOn.Width := FSwitchWidth;
    FSwitchOn.Height := FSwitchHeight;
    ActiveColor := clLime;
    BorderStyle := bsRaised;
  end;
  With FSwitchOff do begin
    Active := false;
    Shape := stRectangle;
    FSwitchOff.Width := FSwitchWidth;
    FSwitchOff.Height := FSwitchHeight;
    InactiveColor := clGray;
    BorderStyle := bsNone;
  end;
end;

destructor TrmCommutator.Destroy;
begin
  FBmpBuffer.free;
  FLedOn.Free;
  FLedOff.Free;
  FSwitchOn.Free;
  FSwitchOff.Free;
  Inherited;
end;

procedure TrmCommutator.Free;
begin
  FBmpBuffer.Free;
  Inherited Free;
end;

procedure TrmCommutator.SetBounds(aLeft, aTop, aWidth, aHeight: Integer);
begin
  Inherited ;
  If FBmpBuffer <> NIL then begin
    FBmpBuffer.Destroy;
    FBmpBuffer := TBitmap.Create;
    FBmpBuffer.Height := aHeight;
    FBmpBuffer.Width := aWidth;
  end;
end;

procedure TrmCommutator.UpdateBmpBuffer;

  function GetSwitchesCountFromSize(var HorzCount, VertCount: integer): integer;
  begin
    if (FOrientation = voHorizontal) then begin
      HorzCount := trunc((width + FSwitchSpacing - (2 * FBorderMargin)) / (FSwitchWidth + FSwitchSpacing));
      VertCount := trunc((height + FSwitchHeight - (2 * FBorderMargin)) / (FSwitchHeight + FSwitchSpacing));
    end else if (FOrientation = voVertical) then begin
      HorzCount := trunc((height + FSwitchSpacing - (2 * FBorderMargin)) / (FSwitchWidth + FSwitchSpacing));
      VertCount := trunc((width + FSwitchSpacing - (2 * FBorderMargin)) / (FSwitchHeight + FSwitchSpacing));
    end;
    result := HorzCount * VertCount;
  end;

var
  Row, line, X1, Y1, X2, Y2, X1Led, Y1Led, X2Led, Y2Led, HorzSwitchesCount, VertSwitchesCount: integer;
  TempSize1, TempSize2: integer;
begin

  // if AutoSize, rmCommutator size is determined by RowCount, LineCount,
  // SwitchLength, SwitchWidth, SwitchSpacing, BorderMargin and ExtremityMargin properties,
  // and cannot be set manually (in conception or execution).
  if AutoSize and not(csReading	in ComponentState) then begin
    TempSize1 := (2 * FBorderMargin) + (FLineCount * FSwitchHeight) + ((FLineCount - 1) * FSwitchSpacing);
    TempSize2 := (2 * FBorderMargin) + (FRowCount * FSwitchWidth) + ((FRowCount - 1) * FSwitchSpacing);
    if round(TempSize1 / 2) <> (TempSize1 / 2) then
      tempSize1 := tempSize1 + 1;
    if (FOrientation = voHorizontal) then begin
      height := TempSize1;
      width := TempSize2;
    end else if (FOrientation = voVertical) then begin
      height := TempSize2;
      width := tempSize1;
    end;
  end;

  // How many elements to draw ?
  // if AutoSize, num of elements are directly obtained by FRowCount and FLineCOunt.
  // else, num of elements are calaculated dependantly of sizes and margins.
  if AutoSize then begin
    HorzSwitchesCount := FRowCount;
    VertSwitchesCount := FLineCount;
  end else begin
    GetSwitchesCountFromSize(HorzSwitchesCount, VertSwitchesCount);
    FRowCount := HorzSwitchesCount;
    FLineCount := VertSwitchesCount;
    if FSwitchIndex > ((FRowCount * FLineCount) - 1) then
      FSwitchIndex := 0;
  end;

  With FBmpBuffer.Canvas do begin

    // background
    Brush.Color := FBGColor;
    Pen.Color := FBGColor;
    Rectangle(0, 0, width, height);
    FLedOn.FRefreshParentAfterChange := false; // avoid loop drawing reaction
    FLedOn.BackGroundColor := FSwitchOn.ActiveColor;
    FLedOn.FRefreshParentAfterChange := true;  // restore normal drawing
    FLedOff.FRefreshParentAfterChange := false; // avoid loop drawing reaction
    FLedOff.BackGroundColor := FSwitchOff.InactiveColor;
    FLedOff.FRefreshParentAfterChange := true;  // restore normal drawing
    FSwitchOn.FRefreshParentAfterChange := false; // avoid loop drawing reaction
    FSwitchOn.BackGroundColor := FSwitchOn.ActiveColor;
    FSwitchOn.FRefreshParentAfterChange := true;  // restore normal drawing
    FSwitchOff.FRefreshParentAfterChange := false; // avoid loop drawing reaction
    FSwitchOff.BackGroundColor := FSwitchOff.InactiveColor;
    FSwitchOff.FRefreshParentAfterChange := true;  // restore normal drawing

    // 3D for commutator border ?
    if FBorderStyle <> bsNone then begin
      if FBorderStyle = bsLowered then
        Pen.Color := clGray
      else
        Pen.Color := clWhite;
      moveTo(0, 0);
      LineTo(0, height - 1);
      moveTo(0, 0);
      LineTo(width - 1, 0);
      if FBorderStyle = bsLowered then
        Pen.Color := clWhite
      else
        Pen.Color := clGray;
      moveTo(0, height - 1);
      LineTo(width - 1, height - 1);
      moveTo(width - 1, height - 1);
      LineTo(width - 1, 0);
    end;

    // for each switches...
    For row := 0 to pred(HorzSwitchesCount) do begin
      for line := 0 to pred(VertSwitchesCount) do begin
        if (FOrientation = voHorizontal) then begin
          X1 := FBorderMargin + ((FSwitchWidth + FSwitchSpacing) * row);
          Y1 := FBorderMargin + ((FSwitchHeight + FSwitchSpacing) * line);
          X2 := X1 + FSwitchWidth;
          Y2 := Y1 + FSwitchHeight;
        end else if (FOrientation = voVertical) then begin
          X1 := FBorderMargin + ((FSwitchHeight + FSwitchSpacing) * line);
          Y1 := FBorderMargin + ((FSwitchWidth + FSwitchSpacing) * row);
          X2 := X1 + FSwitchHeight;
          Y2 := Y1 + FSwitchWidth;
        end;

          // selected switch ?
          if ((row + (Line * FRowCount)) = FSwitchIndex) then begin

            CopyRect(rect(X1, Y1, X2, Y2),
                     FSwitchOn.FBmpBuffer.Canvas,
                     rect(0, 0, FSwitchOn.FBmpBuffer.Width, FSwitchOn.FBmpBuffer.Height));

          end else begin  // switch not selected

            CopyRect(rect(X1, Y1, X2, Y2),
                     FSwitchOff.FBmpBuffer.Canvas,
                     rect(0, 0, FSwitchOff.FBmpBuffer.Width, FSwitchOff.FBmpBuffer.Height));

          end;


        // led on switch's center ?
        if FShowLed then begin

          // selected switch ?
          if ((row + (Line * FRowCount)) = FSwitchIndex) then begin

            X1Led := X1 + trunc((X2 - X1 - FLedOn.Width) / 2);
            Y1Led := Y1 + trunc((Y2 - Y1 - FLedOn.height) / 2);
            X2Led := X1Led + FLedOn.width;
            Y2Led := Y1Led + FLedOn.height;

            CopyRect(rect(X1Led, Y1Led, X2Led, Y2Led),
                     FLedOn.FBmpBuffer.Canvas,
                     rect(0, 0, FLedOn.FBmpBuffer.Width, FLedOn.FBmpBuffer.Height));

          end else begin  // switch not selected
            X1Led := X1 + trunc((X2 - X1 - FLedOff.Width) / 2);
            Y1Led := Y1 + trunc((Y2 - Y1 - FLedOff.height) / 2);
            X2Led := X1Led + FLedOff.width;
            Y2Led := Y1Led + FLedOff.height;

            CopyRect(rect(X1Led, Y1Led, X2Led, Y2Led),
                     FLedOff.FBmpBuffer.Canvas,
                     rect(0, 0, FLedOff.FBmpBuffer.Width, FLedOff.FBmpBuffer.Height));

          end;
        end;
      end;
    end;
  end;
end;

procedure TrmCommutator.Paint;
begin
  UpdateBmpBuffer;
  Canvas.Draw(0, 0, FBmpBuffer);
end;

// when FAllowAllUp set to true, SwitchIndex can take -1 value
// by mouse clicking. Else, if FAllowAllUp set to false,
// SwitchIndex can take -1 value only programmaticaly.
procedure TrmCommutator.SetAllowAllUp(value: boolean);
begin
  if value <> FAllowAllUp then
  begin
    FAllowAllUp := value;
    Paint;
  end;
end;

procedure TrmCommutator.SetShowLed(value: boolean);
begin
  if value <> FShowLed then
  begin
    FShowLed := value;
    Paint;
  end;
end;

procedure TrmCommutator.SetBGColor(value: TColor);
begin
  if value <> FBGColor then
  begin
    FBGColor := value;
    Paint;
  end;
end;

procedure TrmCommutator.SetBorderStyle(value: TrmBorderStyle);
begin
  if value <> FBorderStyle then
  begin
    FBorderStyle := value;
    Paint;
  end;
end;

procedure TrmCommutator.SetOrientation(value: TOrientation);
var
  oldWidthValue: integer;
begin
  if value <> FOrientation then
  begin
    FOrientation := value;
    if (csDesigning in ComponentState) and not(csReading in ComponentState) then begin
      oldWidthValue := width;
      width := height;
      height := oldWidthValue;
    end;
    Paint;
  end;
end;

procedure TrmCommutator.SetBorderMargin(value: integer);
begin
  if value <> FBorderMargin then
  begin
    FBorderMargin := value;
    Paint;
  end;
end;

procedure TrmCommutator.SetAutoSize(value: boolean);
begin
  if value <> FAutoSize then
  begin
    FAutoSize := value;
    Paint;
  end;
end;

procedure TrmCommutator.SetRowCount(value: integer);
begin
  if value <> FRowCount then
  begin
    if value < 1 then value := 1;
    if value > 100 then value := 100;
    FRowCount := value;
    if FSwitchIndex > ((FLineCount * FRowCount) - 1) then
      FSwitchIndex := 0;
    Paint;
  end;
end;

procedure TrmCommutator.SetLineCount(value: integer);
begin
  if value <> FLineCount then
  begin
    if value < 1 then value := 1;
    if value > 100 then value := 100;
    FLineCount := value;
    if FSwitchIndex > ((FLineCount * FRowCount) - 1) then
      FSwitchIndex := 0;
    Paint;
  end;
end;

procedure TrmCommutator.SetSwitchIndex(value: integer);
var
  row, Line: integer;
begin
  if value <> FSwitchIndex then
  begin
    FSwitchIndex := value;
    if FSwitchIndex > ((FLineCount * FRowCount) - 1) then
      FSwitchIndex := 0;

    // refresh selectedRow and SelectedLine variables
    line := 0;
    row := FSwitchIndex + 1;
    While (row > FRowCount) do begin
      dec(row, FRowCount);
      inc(line);
    end;

    SelectedRow := row - 1;
    SelectedLine := line;

    // refresh
    Paint;

    // fire OnChange event if user put code
    if assigned(FOnChange) then
      FOnChange(self);
  end;
end;

procedure TrmCommutator.SetSwitchWidth(value: integer);
begin
  if value <> FSwitchWidth then
  begin
    FSwitchWidth := value;
    Paint;
  end;
end;

procedure TrmCommutator.SetSwitchHeight(value: integer);
begin
  if value <> FSwitchHeight then
  begin
    FSwitchHeight := value;
    Paint;
  end;
end;

procedure TrmCommutator.SetSwitchSpacing(value: integer);
begin
  if value <> FSwitchSpacing then
  begin
    FSwitchSpacing := value;
    Paint;
  end;
end;

procedure TrmCommutator.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button <> mbLeft then exit;
  GetSelectedSwitch(X, Y);
  Paint;
  MouseCapt := true;
end;

procedure TrmCommutator.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button <> mbLeft then exit;
  MouseCapt := false;
end;

procedure TrmCommutator.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if not MouseCapt then exit;
  if (X < 0) or (X > width) or (Y < 0) or (Y > height) then exit;
  GetSelectedSwitch(X, Y);
  Paint;
end;

procedure TrmCommutator.GetSelectedSwitch(X, Y: integer);
var
  row, line, selIndex: integer;
  PayWidth, PayHeight: integer;
begin
  if (FOrientation = voHorizontal) then begin
    if X > (FBorderMargin + ((FRowCount - 1) * FSwitchSpacing) + (FRowCount * FSwitchWidth)) then
      exit;
    row := trunc((X * FRowCount) / (Width - FBorderMargin));
    line := trunc((Y * FLineCount) / (Height - FBorderMargin));
  end else if (FOrientation = voVertical) then begin
    if Y > (FBorderMargin + ((FRowCount - 1) * FSwitchSpacing) + (FRowCount * FSwitchWidth)) then
      exit;
    row := trunc((Y * FRowCount) / (height - FBorderMargin));
    line := trunc((X * FLineCount) / (width - FBorderMargin));
  end;
  selIndex := (row + (Line * FRowCount));
  if ((selIndex = FSwitchIndex) and FAllowAllUp) then begin
    // SelectedRow := -1;
    // SelectedLine := -1;
    SetSwitchIndex(-1);
  end else begin
    // SelectedRow := row;
    // SelectedLine := line;
    SetSwitchIndex(selIndex);
  end;
end;

end.




