unit rmSevenSegDisplay;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, Graphics;

type
  TDisplayType = (dtNormal, dtHalf);
  TDecimalPointType = (dpLeft, dpRight);
  TDecimalType = (dtNone, dtFixed, dtAuto);
  TSignValue = (svNone, svPositive, svNegative);
  TOverFlow = (ofTrunc, ofShowErr);
  TSegmentType = (stPolygon, stLine);

  TrmSevenSegDisplay = class(TGraphicControl)
  private
    { Dclarations prives }
    FBmpBuffer: TBitmap;
    FWidth: integer;
    FHeight: integer;
    FBGColor: TColor;
    FBorderColor: TColor;
    FSegmentOnColor: TColor;
    FSegmentOffColor: TColor;
    FSegmentBorderColor: TColor;
    FAutoSize: boolean;
    FDigitCount: word;
    FDigitWidth: word;
    FDigitHeight: word;
    FDisplayWidth: integer;
    FFillWithZero: boolean;
    FDecimalPoint: TDecimalPointType;
    FOverFlow: TOverFlow;
    FSegmentType: TSegmentType;
    FValueStr: string;
    FValueInt: integer;
    procedure SetBGColor(val: TColor);
    procedure SetBorderColor(val: TColor);
    procedure SetSegmentOnColor(val: TColor);
    procedure SetSegmentOffColor(val: TColor);
    procedure SetSegmentBorderColor(val: TColor);
    procedure SetAutoSize(val: boolean);
    procedure SetDigitCount(val: word);
    procedure SetDigitWidth(val: word);
    procedure SetDigitHeight(val: word);
    procedure SetDisplayWidth(val: integer);
    procedure SetFillWithZero(val: boolean);
    procedure SetDecimalPoint(val: TDecimalPointType);
    procedure SetOverFlow(val: TOverFlow);
    procedure SetSegmentType(val: TSegmentType);
    procedure SetValueStr(val: string);
    procedure SetValueInt(val: integer);
  protected
    { Dclarations protges }
    procedure Paint; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure UpdateBmpBuffer;
  public
    { Dclarations publiques }
    constructor Create(anOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Dclarations publies }
    property BGColor: TColor read FBGColor write SetBGColor;
    property BorderColor: TColor read FBorderColor write SetBorderColor;
    property SegmentOnColor: TColor read FSegmentOnColor write SetSegmentOnColor;
    property SegmentOffColor: TColor read FSegmentOffColor write SetSegmentOffColor;
    property SegmentBorderColor: TColor read FSegmentBorderColor write SetSegmentBorderColor;
    property AutoSize: boolean read FAutoSize write SetAutoSize;
    property DigitCount: word read FDigitCount write SetDigitCount;
    property DigitWidth: word read FDigitWidth write SetDigitWidth;
    property DigitHeight: word read FDigitHeight write SetDigitHeight;
    property DisplayWidth: integer read FDisplayWidth write SetDisplayWidth;
    property FillWithZero: boolean read FFillWithZero write SetFillWithZero;
    property DecimalPoint: TDecimalPointType read FDecimalPoint write SetDecimalPoint;
    property OverFlow: TOverFlow read FOverFlow write SetOverFlow;
    property SegmentType: TSegmentType read FSegmentType write SetSegmentType;
    property ValueStr: string read FValueStr write SetValueStr;
    property ValueInt: integer read FValueInt write SetValueInt;
    { Standard properties }
    property Hint;
    property ShowHint;
    { Standard events }
    property OnClick;
    property OnDblClick;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('OptoElec', [TrmSevenSegDisplay]);
end;

{ TrmSevenSegDisplay }

constructor TrmSevenSegDisplay.Create(anOwner: TComponent);
begin
  Inherited Create(anOwner);
  Height := 53;
  width := 160;
  FAutoSize := true;
  FDigitCount := 4;
  FDigitWidth := 40;
  FDigitHeight := 53;
  FDisplayWidth := 160;
  FBmpBuffer := TBitmap.Create;
  FBGColor := clBlack;
  FDecimalPoint := dpRight;
  FSegmentOnColor := clRed;
  FSegmentOffColor := clMaroon;
  FSegmentBorderColor := clGray;
  FBorderColor := clGray;
  FOverFlow := ofShowErr;
  FSegmentType := stLine;
  FValueInt := 9999;
  FValueStr := '9999';
  //AdjustSizes;
end;

destructor TrmSevenSegDisplay.Destroy;
begin
  If FBmpBuffer <> NIL then FBmpBuffer.free;
  Inherited Destroy;
end;

procedure TrmSevenSegDisplay.Paint;
begin
  UpdateBmpBuffer;
  Canvas.Draw(0, 0, FBmpBuffer);
end;

procedure TrmSevenSegDisplay.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  Inherited ;
  If FBmpBuffer <> NIL then begin
    FBmpBuffer.Destroy;
    FBmpBuffer := TBitmap.Create;
    FBmpBuffer.Height := aHeight;
    FBmpBuffer.Width := aWidth;
  end;
end;

procedure TrmSevenSegDisplay.UpdateBmpBuffer;

  procedure DrawBackGround(count: word);
  begin
    With FBmpBuffer.Canvas do begin
      brush.color := FBGColor;
      pen.color := FBorderColor;
      pen.width := 1;
      rectangle(count * FDigitWidth, 0, FDigitWidth * (count + 1), height);
    end;
  end;

  procedure DrawDigit(count: word; c: string);

    procedure DrawSegments(count: word; a, b, c, d, e, f, g, dpL, dpR: boolean);

      procedure SetColors(lit, isPolygon: boolean);
      begin
        with FBmpBuffer.canvas do begin
          if (lit and isPolygon) then begin
            pen.Color := FSegmentBorderColor;
            brush.Color := FSegmentOnColor;
          end else if ((not lit) and isPolygon) then begin
            pen.Color := FSegmentBorderColor;
            brush.Color := FSegmentOffColor;
          end else if (lit and (not isPolygon)) then begin
            pen.Color := FSegmentOnColor;
            brush.Color := FSegmentOnColor;
          end else if ((not lit) and (not isPolygon)) then begin
            pen.Color := FSegmentOffColor;
            brush.Color := FSegmentOffColor;
          end;
        end;
      end;

    var
      HorzOffset: word;
      a1x, a1y, a2x, a2y, a3x, a3y, a4x, a4y,
      b1x, b1y, b2x, b2y, b3x, b3y, b4x, b4y,
      c1x, c1y, c2x, c2y, c3x, c3y, c4x, c4y,
      d1x, d1y, d2x, d2y, d3x, d3y, d4x, d4y,
      e1x, e1y, e2x, e2y, e3x, e3y, e4x, e4y,
      f1x, f1y, f2x, f2y, f3x, f3y, f4x, f4y,
      g1x, g1y, g2x, g2y, g3x, g3y, g4x, g4y,
      dpL1x, dpL1y, dpL2x, dpL2y, dpR1x, dpR1y, dpR2x, dpR2y: word;
    begin
      HorzOffset := count * FDigitWidth;
      a1x := HorzOffset + trunc(FDigitWidth / 100 * 43.50);
      a1y := trunc(Height / 100 * 17.91);
      a2x := HorzOffset + trunc(FDigitWidth / 100 * 77.79);
      a2y := trunc(Height / 100 * 17.91);
      a3x := HorzOffset + trunc(FDigitWidth / 100 * 77.79);
      a3y := trunc(Height / 100 * 23.11);
      a4x := HorzOffset + trunc(FDigitWidth / 100 * 43.50);
      a4y := trunc(Height / 100 * 23.11);
      b1x := HorzOffset + trunc(FDigitWidth / 100 * 80.58);
      b1y := trunc(Height / 100 * 20.26);
      b2x := HorzOffset + trunc(FDigitWidth / 100 * 89.22);
      b2y := trunc(Height / 100 * 20.26);
      b3x := HorzOffset + trunc(FDigitWidth / 100 * 81.60);
      b3y := trunc(Height / 100 * 47.69);
      b4x := HorzOffset + trunc(FDigitWidth / 100 * 72.90);
      b4y := trunc(Height / 100 * 47.69);
      c1x := HorzOffset + trunc(FDigitWidth / 100 * 72.20);
      c1y := trunc(Height / 100 * 49.91);
      c2x := HorzOffset + trunc(FDigitWidth / 100 * 80.77);
      c2y := trunc(Height / 100 * 49.91);
      c3x := HorzOffset + trunc(FDigitWidth / 100 * 73.28);
      c3y := trunc(Height / 100 * 76.83);
      c4x := HorzOffset + trunc(FDigitWidth / 100 * 64.77);
      c4y := trunc(Height / 100 * 76.83);
      d1x := HorzOffset + trunc(FDigitWidth / 100 * 28.26);
      d1y := trunc(Height / 100 * 74.10);
      d2x := HorzOffset + trunc(FDigitWidth / 100 * 62.61);
      d2y := trunc(Height / 100 * 74.10);
      d3x := HorzOffset + trunc(FDigitWidth / 100 * 62.61);
      d3y := trunc(Height / 100 * 79.44);
      d4x := HorzOffset + trunc(FDigitWidth / 100 * 28.26);
      d4y := trunc(Height / 100 * 79.44);
      e1x := HorzOffset + trunc(FDigitWidth / 100 * 25.72);
      e1y := trunc(Height / 100 * 49.91);
      e2x := HorzOffset + trunc(FDigitWidth / 100 * 33.60);
      e2y := trunc(Height / 100 * 49.91);
      e3x := HorzOffset + trunc(FDigitWidth / 100 * 25.59);
      e3y := trunc(Height / 100 * 76.83);
      e4x := HorzOffset + trunc(FDigitWidth / 100 * 17.65);
      e4y := trunc(Height / 100 * 76.83);
      f1x := HorzOffset + trunc(FDigitWidth / 100 * 33.53);
      f1y := trunc(Height / 100 * 20.26);
      f2x := HorzOffset + trunc(FDigitWidth / 100 * 41.40);
      f2y := trunc(Height / 100 * 20.26);
      f3x := HorzOffset + trunc(FDigitWidth / 100 * 33.85);
      f3y := trunc(Height / 100 * 47.69);
      f4x := HorzOffset + trunc(FDigitWidth / 100 * 25.78);
      f4y := trunc(Height / 100 * 47.69);
      g1x := HorzOffset + trunc(FDigitWidth / 100 * 36.00);
      g1y := trunc(Height / 100 * 46.23);
      g2x := HorzOffset + trunc(FDigitWidth / 100 * 69.66);
      g2y := trunc(Height / 100 * 46.23);
      g3x := HorzOffset + trunc(FDigitWidth / 100 * 69.66);
      g3y := trunc(Height / 100 * 51.24);
      g4x := HorzOffset + trunc(FDigitWidth / 100 * 36.00);
      g4y := trunc(Height / 100 * 51.24);
      dpL1x := HorzOffset + trunc(FDigitWidth / 100 * 9.00);
      dpL1y := trunc(Height / 100 * 74.00);
      dpL2x := HorzOffset + trunc(FDigitWidth / 100 * 15.00);
      dpL2y := trunc(Height / 100 * 80.00);
      dpR1x := HorzOffset + trunc(FDigitWidth / 100 * 83.00);
      dpR1y := trunc(Height / 100 * 74.00);
      dpR2x := HorzOffset + trunc(FDigitWidth / 100 * 91.00);
      dpR2y := trunc(Height / 100 * 80.00);

      With FBmpBuffer.Canvas do begin

        // segments, polygon for large size, line for small line (manual choice)
        if FSegmentType = stPolygon then begin
          pen.width := 1;
          SetColors(a, true);
          polygon([Point(a1x, a1y), Point(a2x, a2y), Point(a3x, a3y), Point(a4x, a4y)]);
          SetColors(b, true);
          polygon([Point(b1x, b1y), Point(b2x, b2y), Point(b3x, b3y), Point(b4x, b4y)]);
          SetColors(c, true);
          polygon([Point(c1x, c1y), Point(c2x, c2y), Point(c3x, c3y), Point(c4x, c4y)]);
          SetColors(d, true);
          polygon([Point(d1x, d1y), Point(d2x, d2y), Point(d3x, d3y), Point(d4x, d4y)]);
          SetColors(e, true);
          polygon([Point(e1x, e1y), Point(e2x, e2y), Point(e3x, e3y), Point(e4x, e4y)]);
          SetColors(f, true);
          polygon([Point(f1x, f1y), Point(f2x, f2y), Point(f3x, f3y), Point(f4x, f4y)]);
          SetColors(g, true);
          polygon([Point(g1x, g1y), Point(g2x, g2y), Point(g3x, g3y), Point(g4x, g4y)]);
        end else if FSegmentType = stLine then begin
          pen.width := 2;
          SetColors(a, false);
          moveTo(a1x, a1y + trunc((a4y - a1y) / 2));
          lineTo(a2x, a2y + trunc((a3y - a2y) / 2));
          SetColors(b, false);
          moveTo(b1x + trunc((b2x - b1x) / 2), b1y);
          lineTo(b4x + trunc((b3x - b4x) / 2), b3y);
          SetColors(c, false);
          moveTo(c1x + trunc((c2x - c1x) / 2), c1y);
          lineTo(c4x + trunc((c3x - c4x) / 2), c3y);
          SetColors(d, false);
          moveTo(d1x, d1y + trunc((d4y - d1y) / 2));
          lineTo(d2x, d2y + trunc((d3y - d2y) / 2));
          SetColors(e, false);
          moveTo(e1x + trunc((e2x - e1x) / 2), e1y);
          lineTo(e4x + trunc((e3x - e4x) / 2), e3y);
          SetColors(f, false);
          moveTo(f1x + trunc((f2x - f1x) / 2), f1y);
          lineTo(f4x + trunc((f3x - f4x) / 2), f3y);
          SetColors(g, false);
          moveTo(g1x, g1y + trunc((g4y - g1y) / 2));
          lineTo(g2x, g2y + trunc((g3y - g2y) / 2));
        end;

        // decimal point, left or rigth
        if (FDecimalPoint = dpLeft) then begin
          setColors(dpL, false);
          ellipse(dpL1x, dpL1y, dpL2x, dpL2y);
        end else if (FDecimalPoint = dpRight) then begin
          setColors(dpR, false);
          ellipse(dpR1x, dpR1y, dpR2x, dpR2y);
        end;
      end;
    end;

  var
    dpL, dpR, SixNineQueue: boolean;
    DPPos: word;
  begin
    DPPos := pos('.', FValueStr);                      // search for point in FValue
    if DPPos = 0 then DPPos := pos(',', FValueStr);    // if not point, search for comma
    if (DPPos > 0) then begin
      dpL := Length(FValueStr) - DPPos = FDigitCount - count;
      dpR := Length(FValueStr) - DPPos + 1 = FDigitCount - count;
    end else begin
      dpL := false;
      dpR := false;
    end;
    SixNineQueue := true;  // if you want you can add a "que" property for "6" and "9" digits
    With FBmpBuffer.Canvas do begin
      // textOut(count * FDigitWidth + 6, 6, c); // for test
      if c = '0' then DrawSegments(count, true, true, true, true, true, true, false, dpL, dpR)
      else if c = '1' then DrawSegments(count, false, true, true, false, false, false, false, dpL, dpR)
      else if c = '2' then DrawSegments(count, true, true, false, true, true, false, true, dpL, dpR)
      else if c = '3' then DrawSegments(count, true, true, true, true, false, false, true, dpL, dpR)
      else if c = '4' then DrawSegments(count, false, true, true, false, false, true, true, dpL, dpR)
      else if c = '5' then DrawSegments(count, true, false, true, true, false, true, true, dpL, dpR)
      else if c = '6' then DrawSegments(count, SixNineQueue, false, true, true, true, true, true, dpL, dpR)
      else if c = '7' then DrawSegments(count, true, true, true, false, false, false, false, dpL, dpR)
      else if c = '8' then DrawSegments(count, true, true, true, true, true, true, true, dpL, dpR)
      else if c = '9' then DrawSegments(count, true, true, true, SixNineQueue, false, true, true, dpL, dpR)
      else if c = '-' then DrawSegments(count, false, false, false, false, false, false, true, dpL, dpR)
      else if c = 'E' then DrawSegments(count, true, false, false, true, true, true, true, false, false)
      else if c = '' then DrawSegments(count, false, false, false, false, false, false, false, false, false);
    end;
  end;

var
  i, count, start: word;
  sVal: string;
  DPFound: boolean;
  OverFlow: boolean;
  c: string[1];
begin
  // adjust digit relative properties if necessary
  if FAutoSize then begin
    FDigitCount := trunc(Width / FDigitWidth);
    FDigitHeight := trunc(FDigitWidth * 4 / 3);
    height := FDigitHeight;
    if width <> trunc(FDigitCount * FDigitWidth) then
      width := trunc(FDigitCount * FDigitWidth);
  end else begin
    FDigitWidth := trunc(width / FDigitCount);
  end;
  FDisplayWidth := width;

  // search for decimal point
  sVal := FValueStr;
  DPFound := (pos('.', sVal) > 0) or (pos(',', sVal) > 0);
  OverFlow := (DPFound and (length(sval) > (FDigitCount + 1)) and (FOverFlow = ofShowErr)) or
              ((not DPFound) and (length(sval) > FDigitCount) and (FOverFlow = ofShowErr));
  if OverFlow then begin
    for i := FDigitCount - 1 downTo 0 do begin
      DrawBackGround(i);
      DrawDigit(i, 'E');
    end;
  end else begin
    if DPFound then start := FDigitCount else start := FDigitCount - 1;
    count := FDigitCount - 1;
    for i := start downTo 0 do begin
      DrawBackGround(Count);
      c := copy(sVal, length(sVal), 1);
      delete(sVal, length(sVal), 1);
      if ((c <> '') and (c <> '.') and (c <> ',')) then DrawDigit(Count, c);
      if ((c = '') and FFillWithZero) then DrawDigit(Count, '0')
      else if ((c = '') and (not FFillWithZero)) then DrawDigit(Count, '');
      if ((c <> '.') and (c <> ',')) then dec(Count);
    end;
  end;
end;

procedure TrmSevenSegDisplay.SetBGColor(val: TColor);
begin
  if val <> FBGColor then begin
    FBGColor := val;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetBorderColor(val: TColor);
begin
  if val <> FBorderColor then begin
    FBorderColor := val;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetSegmentOnColor(val: TColor);
begin
  if val <> FSegmentOnColor then begin
    FSegmentOnColor := val;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetSegmentOffColor(val: TColor);
begin
  if val <> FSegmentOffColor then begin
    FSegmentOffColor := val;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetSegmentBorderColor(val: TColor);
begin
  if val <> FSegmentBorderColor then begin
    FSegmentBorderColor := val;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetDigitCount(val: word);
begin
  if val <> FDigitCount then begin
    FDigitCount := val;
    if FAutoSize then begin
      FDisplayWidth := FDigitCount * FDigitWidth;
      FDigitHeight := trunc(FDigitWidth * 4 / 3);
      width := FDisplayWidth;
    end else begin
      FDigitWidth := trunc(FDisplayWidth / FDigitCount);
    end;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetValueStr(val: string);
begin
  if val <> FValueStr then begin
    FValueStr := val;
    try
      FValueInt := StrToInt(FValueStr);
    except
      FValueInt := 0;
      FValueStr := '0'
    end;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetValueInt(val: integer);
begin
  if val <> FValueInt then begin
    FValueInt := val;
    try
      FValueStr := IntToStr(FValueInt);
    except
      FValueStr := '0';
      FValueInt := 0;
    end;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetDigitWidth(val: word);
begin
  if val <> FDigitWidth then begin
    FDigitWidth := val;
    if FAutoSize then begin
      FDisplayWidth := FDigitCount * FDigitWidth;
      FDigitHeight := trunc(FDigitWidth * 4 / 3);
      width := FDisplayWidth;
    end else begin
      FDigitCount := trunc(FDisplayWidth / FDigitWidth);
    end;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetDigitHeight(val: word);
begin
  if val <> FDigitHeight then begin
    FDigitHeight := val;
    if FAutoSize then begin
      FDigitWidth := trunc(FDigitHeight * 3 / 4);
      FDisplayWidth := FDigitCount * FDigitWidth;
      width := FDisplayWidth;
    end else begin
      FDigitWidth := trunc(FDisplayWidth / FDigitCount);
    end;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetDisplayWidth(val: integer);
begin
  if val <> FDisplayWidth then begin
    FDisplayWidth := val;
    width := FDisplayWidth;
    if FAutoSize then begin
      FDigitCount := trunc(FDisplayWidth / FDigitWidth);
      FDigitHeight := trunc(FDigitWidth * 4 / 3);
    end else begin
      FDigitWidth := trunc(FDisplayWidth / FDigitCount);
    end;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetAutoSize(val: boolean);
begin
  if val <> FAutoSize then begin
    FAutoSize := val;
    if FAutoSize then begin
      FDigitHeight := trunc(FDigitWidth * 4 / 3);
      Height := FDigitHeight;
      FDisplayWidth := FDigitCount * FDigitWidth;
      width := FDisplayWidth;
    end;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetFillWithZero(val: boolean);
begin
  if val <> FFillWithZero then begin
    FFillWithZero := val;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetDecimalPoint(val: TDecimalPointType);
begin
  if val <> FDecimalPoint then begin
    FDecimalPoint := val;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetOverFlow(val: TOverFlow);
begin
  if val <> FOverFlow then begin
    FOverFlow := val;
    paint;
  end;
end;

procedure TrmSevenSegDisplay.SetSegmentType(val: TSegmentType);
begin
  if val <> FSegmentType then begin
    FSegmentType := val;
    paint;
  end;
end;


end.
